;; bytecomp-patch.el

(require 'byte-opt)

;; Compile a function that accepts one or more args and is right-associative.
;; We do it by left-associativity so that the operations
;; are done in the same order as in interpreted code.
;; We treat the one-arg case, as in (+ x), like (+ x 0).
;; in order to convert markers to numbers, and trigger expected errors.
(defun byte-compile-associative (form)
  (if (cdr form)
      (let ((opcode (get (car form) 'byte-opcode))
	    args)
	(if (and (< 3 (length form))
		 (memq opcode (list (get '+ 'byte-opcode)
				    (get '* 'byte-opcode))))
	    ;; (+ a b ...), (* a b ...)
	    (byte-compile-normal-call form)
	  (setq args (copy-sequence (cdr form)))
	  (byte-compile-form (car args))
	  (setq args (cdr args))
	  (or args (setq args '(0)
			 opcode (get '+ 'byte-opcode)))
	  (dolist (arg args)
	    (byte-compile-form arg)
	    (byte-compile-out opcode 0))))
    (byte-compile-constant (eval form))))


(defun byte-compile-minus (form)
  (let ((len (length form)))
    (cond
     ((= 1 len) (byte-compile-constant 0))
     ((= 2 len)
      (byte-compile-form (cadr form))
      (byte-compile-out 'byte-negate 0))
     ((= 3 len) 
      (byte-compile-form (nth 1 form))
      (byte-compile-form (nth 2 form))
      (byte-compile-out 'byte-diff 0))
     (t (byte-compile-normal-call form)))))


(defun byte-compile-quo (form)
  (let ((len (length form)))
    (cond ((<= len 2)
	   (byte-compile-subr-wrong-args form "2 or more"))
	  ((= len 3)
	   (byte-compile-two-args form))
	  (t
	   (byte-compile-normal-call form)))))


;; byte-opt

(when nil
  (let ((side-effect-free-fns
	 '(% * + - / /= 1+ 1- < <= = > >= abs acos append aref ash asin atan
	     assoc assq
	     boundp
;;;+++
	     buffer-chars-modified-tick
;;;---
	     buffer-file-name buffer-local-variables buffer-modified-p
;;;+++
	     buffer-modified-tick
;;;---
	     buffer-substring byte-code-function-p
;;;+++
	     byte-to-position
;;;---	     
	     capitalize car-less-than-car car cdr ceiling char-after char-before
;;;+++
	     char-bytes char-direction
;;;---
	     char-equal char-to-string char-width
;;;+++
	     color-distance
;;;---
	     compare-strings concat 
;;;+++
	     constrain-to-field
;;;---
	     coordinates-in-window-p
	     copy-alist copy-sequence copy-marker cos 
;;;+++
	     count-lines		;not subr
	     decode-big5-char decode-char decode-sjis-char
;;;---
	     decode-time default-boundp default-value documentation downcase
	     elt
;;;+++
	     encode-big5-char encode-char encode-sjis-char
;;;---
	     exp expt encode-time error-message-string
	     fboundp fceiling featurep ffloor
;;;+++
	     field-beginning field-end
;;;---
	     file-directory-p file-exists-p file-locked-p file-name-absolute-p
	     file-newer-than-file-p file-readable-p file-symlink-p file-writable-p
	     float float-time floor format format-time-string frame-visible-p
	     fround ftruncate
	     get gethash get-buffer get-buffer-window getenv get-file-buffer
	     hash-table-count
;;;+++
	     hash-table-size
;;;---
	     int-to-string intern-soft
	     keymap-parent
	     length
;;;+++
	     line-beginning-position line-end-position
;;;---
	     local-variable-if-set-p local-variable-p log log10 logand
	     logb logior lognot logxor lsh
	     make-list make-string make-symbol
	     marker-buffer
;;;+++
	     marker-position match-beginning match-end
;;;---
	     max member memq min mod multibyte-char-to-unibyte
;;;+++
	     next-char-property-change
	     next-overlay-change
	     next-property-change
	     next-single-char-property-change
	     next-single-property-change
;;;---
	     next-window nth nthcdr number-to-string
	     parse-colon-path plist-get plist-member
;;;+++
	     position-bytes
;;;---
	     prefix-numeric-value 
;;;+++
	     previous-char-property-change
	     previous-overlay-change
	     previous-property-change
	     previous-single-char-property-change
	     previous-single-property-change
;;;---
	     previous-window prin1-to-string propertize
;;;+++ ???????
	     radians-to-degrees		;not subr
;;;---
	     rassq rassoc read-from-string regexp-quote
	     region-beginning region-end reverse round
	     sin sqrt string string< string=
;;;+++
	     string-bytes
;;;---
	     string-equal string-lessp string-to-char
	     string-to-int string-to-number 
;;;+++
	     string-width
;;;---
	     substring sxhash symbol-function
	     symbol-name symbol-plist symbol-value string-make-unibyte
	     string-make-multibyte string-as-multibyte string-as-unibyte
;;;+++
	     sxhash
;;;---
	     tan 
;;;+++
	     text-property-any text-property-not-all
;;;---
	     truncate
	     unibyte-char-to-multibyte upcase user-full-name
	     user-login-name user-original-login-name user-variable-p
	     vconcat
	     window-buffer window-dedicated-p
;;;+++
	     window-end
;;;---
	     window-edges window-height
	     window-hscroll window-minibuffer-p
;;;+++
	     window-start window-text-height window-vscroll
;;;---
	     window-width
	     zerop))
	(side-effect-and-error-free-fns
	 '(arrayp
	   atom
	   bobp bolp bool-vector-p
	   buffer-end buffer-list buffer-size buffer-string bufferp
;;;+++
	   byteorder
;;;---	   
	   car-safe case-table-p cdr-safe
	   char-or-string-p commandp cons consp
	   current-buffer 
;;;+++
	   current-column
;;;---
	   current-global-map current-indentation
	   current-local-map current-minor-mode-maps current-time
	   current-time-string current-time-zone
;;;+++
	   emacs-pid
;;;---
	   eobp eolp eq equal eventp
	   floatp following-char framep
;;;+++
	   gap-position gap-size
;;;---
	   get-largest-window get-lru-window
	   hash-table-p
	   identity ignore integerp integer-or-marker-p interactive-p
	   invocation-directory invocation-name
	   keymapp
	   line-beginning-position line-end-position list listp
	   make-marker mark mark-marker markerp memory-limit
;;;+++
	   minibuffer-depth minibuffer-prompt-end
;;;---
	   minibuffer-window
	   mouse-movement-p
	   natnump nlistp not null number-or-marker-p numberp
	   one-window-p overlayp
	   point point-marker point-min point-max preceding-char processp
	   recent-keys recursion-depth
	   safe-length selected-frame selected-window sequencep
	   standard-case-table standard-syntax-table stringp subrp symbolp
	   syntax-table syntax-table-p
	   this-command-keys this-command-keys-vector this-single-command-keys
	   this-single-command-raw-keys
	   user-real-login-name user-real-uid user-uid
	   vector vectorp visible-frame-list
	   wholenump window-configuration-p window-live-p windowp)))
    (while side-effect-free-fns
      (put (car side-effect-free-fns) 'side-effect-free t)
      (setq side-effect-free-fns (cdr side-effect-free-fns)))
    (while side-effect-and-error-free-fns
      (put (car side-effect-and-error-free-fns) 'side-effect-free 'error-free)
      (setq side-effect-and-error-free-fns (cdr side-effect-and-error-free-fns)))
    nil)

;;; pure functions are side-effect free functions whose values depend
;;; only on their arguments. For these functions, calls with constant
;;; arguments can be evaluated at compile time. This may shift run time
;;; errors to compile time.
  (let ((pure-fns
       '(concat symbol-name regexp-opt regexp-quote string-to-syntax
;;; make-string
		)))
  (while pure-fns
    (put (car pure-fns) 'pure t)
    (setq pure-fns (cdr pure-fns)))
  nil)
  )

(defconst byte-optimize-float-functions
  '(float 
    fceiling ffloor fround ftruncate
    float-time
    sin cos tan acos asin atan
    exp expt log log10 sqrt
    ;; radians-to-degrees
    )
  "List of functions that return a float."
  )

(defconst byte-optimize-marker-functions
  '(copy-marker
    make-marker
    ;; mark
    mark-marker
    point-marker
    point-max-marker
    point-min-marker
    set-marker)
  "List of functions that return a marker."
  )

(defconst byte-optimize-number-functions
  '(*
    + - / 1+ 1- abs
    ;; count-lines
    max min
    string-to-int
    string-to-number
    user-real-uid user-uid)
  "List of functions that return a number (intgeger, float)."
  )

(defconst byte-optimize-integer-functions
  '(%
    ash
   ;; buffer-end
    buffer-chars-modified-tick
    buffer-modified-tick
    buffer-size
    byte-to-position
    byteorder
    ceiling
    char-after char-before
    char-bytes
    char-direction
    ;; char-resolve-modifiers
    char-width
    ;; charset-id-internal
    color-distance
    constrain-to-field
    ;; count-lines
    current-column
    current-indentation
    ;; decode-big5-char
    ;; decode-char
    ;; decode-sjis-char
    ;; encode-big5-char
    ;; encode-char
    ;; encode-sjis-char
    emacs-pid
    field-beginning field-end
    floor
    following-char
    forward-line
    forward-point
    gap-position gap-size
    get-file-char
    goto-char
    hash-table-count
    hash-table-size
    indent-to
    length
    line-beginning-position line-end-position
    logand logb logior lognot logxor lsh
    ;; make-char
    marker-position			;nil
    match-beginning match-end		;nil
    ;; max-char
    memory-limit
    minibuffer-depth
    minibuffer-prompt-end
    mod
    move-to-column
    move-to-window-line
    multibyte-char-to-unibyte
    next-char-property-change
    next-overlay-change
    next-property-change
    next-single-char-property-change
    next-single-property-change
    point point-min point-max
    position-bytes
    posix-search-backward posix-search-forward
    preceding-char
    prefix-numeric-value
    previous-char-property-change
    previous-overlay-change
    previous-property-change
    previous-single-char-property-change
    previous-single-property-change
    ;; primary-charset
    random
    read-char read-char-exclusive
    recursion-depth
    re-search-backward re-search-forward
    region-beginning region-end
    round
    safe-length
    scroll-left scroll-right
    search-backward search-forward	;nil
    set-window-hscroll
    set-window-vscroll
    set-window-point
    set-window-redisplay-end-trigger	;nil
    set-window-start
    skip-chars-backward skip-chars-forward
    skip-syntax-backward skip-syntax-forward
    string-bytes
    string-to-char
    ;; string-to-int 
    string-width
    sxhash
    text-property-any text-property-not-all ;nil
    truncate
    unibyte-char-to-multibyte
    ;; user-real-uid user-uid
    vertical-motion
    window-end
    window-height window-hscroll
    window-start
    window-text-height
    window-vscroll
    window-width
    word-search-backward word-search-forward
    write-char
    )
  "List of functions that return a non-marker integer.
Some may return nil in some case."
  )


(defun byte-optimize-nonmarker-p (arg)
  ;; Return t if ARG's value is non-marker number.
  (or (numberp arg)
      (and (consp arg)
	   (or (memq (car arg) byte-optimize-integer-functions)
	       (memq (car arg) byte-optimize-number-functions)
	       (memq (car arg) byte-optimize-float-functions))
	   t)))


(defun byte-optimize-integer-or-marker-p (arg)
  ;; Return t if ARG's value is not an integer nor a marker.
  (or (integer-or-marker-p arg)
      (and (consp arg)
	   (or (memq (car arg) byte-optimize-integer-functions)
	       (memq (car arg) byte-optimize-marker-functions))
	   t)))


(defun byte-optimize-floatp (arg)
  ;; Return t if ARG's value is a float.
  (or (floatp arg)
      (and (consp arg)
	   (or (and (memq (car arg) byte-optimize-float-functions) t)
	       (and (memq (car arg) '(+ - * / 1+ 1- abs max min))
		    (byte-optimize-args-float-p (cdr arg)))))))


(defun byte-optimize-side-effect-free-p (arg)
  ;; Return t if ARG is side-effect free at runtime.
  (or (numberp arg)
      (symbolp arg)
      (and (consp arg)
	   (get (car arg) 'side-effect-free)
	   (null (memq nil
		       (mapcar #'byte-optimize-side-effect-free-p
			       (cdr arg)))))))


(defun byte-optimize-args-float-p (args)
  ;; Return t if some of ARGS is found at bytecompile time to
  ;; return a float number.
  (while (and args
	      (null (byte-optimize-floatp (car args))))
    (setq args (cdr args)))
  (and args t))


(defun byte-optimize-args-integer-or-marker-p (args)
  ;; Return t if any of ARGS is found at bytecompile time to return
  ;; integer or marker. In other words, return t when args never
  ;; return floats.
  (while (and args
	      (byte-optimize-integer-or-marker-p (car args)))
    (setq args (cdr args)))
  (null args))

(defsubst byte-compile-butlast (form)
  (nreverse (cdr (reverse form))))

(defun byte-optimize-arith-remove-constant (form start int)
  ;; Remove all FORM's constants INT from number START.
  ;; If FORM's value seems float, remove float constants also.
  (let ((rest (nthcdr (1- start) form)))
    (when (or (memq int (cdr rest)) (memql (float int) (cdr rest)))
      ;; remove integer anyway.
      (setq form (copy-sequence form)
	    rest (nthcdr (1- start) form))
      (setcdr rest (delq int (cdr rest)))
      ;; remove float.
      (when (memql (float int) (cdr rest))
	(setcdr rest (delete (float int) (cdr rest)))
	(setq rest (cdr form))
	;; ensure form float.
	(cond
	 ;; some other argument is float.
	 ((byte-optimize-args-float-p rest))
	 ;; one integer to float.
	 ((memq t (mapcar #'integerp rest))
	  (while (null (integerp (car rest))) (setq rest (cdr rest)))
	  (setcar rest (float (car rest))))
	 ;; give up and append one float to the last.
	 (t
	  (setq form (nconc form (list (float int))))))))
    form))

    
(defun byte-optimize-arith-preprocess (form)
  ;; Prepare or preprocess for arithmetic operations.
  (when (or (cddr form)
	    (and (eq (car form) '*) (null (or (memq 0 (cdr form))
					      (memql 0.0 (cdr form)))))
	    (and (eq (car form) '/) (null (or (member (cadr form) '(0 0.0))
					      (memq 0 (cddr form))
					      (memql 0.0 (cddr form))))))
    (let ((op (car form))
	  (args (cdr form)))
      ;; pre-calculate first part of constant args without integer
      ;; overflow nor truncation.
      (when (and (numberp (car args)) (numberp (cadr args)))
	(setq args (copy-sequence args)
	      form (cons op args))
	;; float oparation
	(if (byte-optimize-args-float-p args)
	    (setcar args (float (car args))))
	(while (and (numberp (car args)) (numberp (cadr args)))
	  (let ((result (funcall op (car args) (cadr args))))
	    (if (byte-optimize-approx-equal
		 result
		 (funcall op (float (car args)) (cadr args)))
		(progn
		  (setcar args result)
		  (setcdr args (cddr args)))
	      ;; exit on overflow or truncation.
	      (setq args nil)))))
      ;; remove noisy constants from form
      (setq form (byte-optimize-arith-remove-constant
		  form
		  (if (memq op '(+ *)) 1 2)
		  (if (memq op '(+ -)) 0 1)))
      ;; -1 in '*','/' float operation
      (when (memq op '(* /))
	(setq args (if (eq op '*) (cdr form) (cddr form)))
	(let ((count (apply '+ (mapcar (lambda (x)
					 (if (or (eq x -1) (eql x -1.0))
					     1
					   0))
				       args))))
	  ;; -1.0 or (-1 and some float argument)
	  (when (and (> count 0)
		     (or (memql -1.0 args)
			 (and (memq -1 args)
			      (byte-optimize-args-float-p (cdr form)))))
	    (setq args (delete -1.0 (delq -1 (copy-sequence args))))
	    (cond
	     ((eq '* op)
	      (setq form (cons op args)))
	     ((eq '/ op)
	      (setq form (nconc (list op (cadr form)) args)
		    args (cdr form))))
	    (when (eq 1 (logand 1 count))
	      (cond
	       ((memq t (mapcar #'numberp args))
		(while (null (numberp (car args))) (setq args (cdr args)))
		(setcar args (- (float (car args)))))
	       ((byte-optimize-args-float-p args)
		(setq form (nconc form (list -1))))
	       ((setq form (nconc form (list -1.0)))))))))
      ;; correct broken expression
      (or (cddr form)
	  (cond
	   ;; (- x) --> (- x 0)
	   ((eq '- (car form))
	    (setq form (nconc form (list 0))))
	   ;; (/ x) --> (/ x 1)
	   ((eq '/ (car form))
	    (setq form (nconc form (list 1))))))))
  form)


(defun byte-optimize-force-binary-operation (form)
  ;; Make expression in FORM to binary operation.
  (when (< 3 (length form))
    (let ((op (car form))
	  (args (nthcdr 3 form)))
      (setq form (list op (nth 1 form) (nth 2 form)))
      (while args
	(setq form (list op form (car args))
	      args (cdr args)))))
  form)


(defun byte-optimize-force-binary-operation-if (form &optional limit)
  (if (and (< 3 (safe-length form))
	   (or (null limit)
	       (<= (length form) limit))
	   (or (byte-optimize-floatp (nth 1 form))
	       (byte-optimize-floatp (nth 2 form))
	       (byte-optimize-args-integer-or-marker-p (cdr form))))
      (byte-optimize-force-binary-operation form)
    form))


(defun byte-optimize-arith-one (form op)
  ;; Optimization for add/sbtract operations that can use `1+', `1-'.
  (let* ((len (length form))
	 (first (nth 1 form))
	 (last (nth (1- len) form)))
    (cond
     ;; (op FIRST LAST)
     ((= 3 len)
      (cond
       ((or (eq last 1)
	    (and (eql last 1.0) (byte-optimize-floatp first)))
	(setq form (list (if (eq '+ op) '1+ '1-) first)))
       ((or (eq last -1)
	    (and (eql last -1.0) (byte-optimize-floatp first)))
	(setq form (list (if (eq '+ op) '1- '1+) first)))
       ((or (eq first 1)
	    (and (eql first 1.0) (byte-optimize-floatp last)))
	(setq form (list (if (eq '+ op) '1+ '1-) last)))
       ((or (eq first -1)
	    (and (eql first -1.0) (byte-optimize-floatp last)))
	(setq form (list (if (eq '+ op) '1- '1+) last)))))
     ;; (op FIRST x LAST)
     ((= 4 len)
      (cond
       ((or (eq last 1)
	    (and (eql last 1.0)
		 (or (byte-optimize-floatp first)
		     (byte-optimize-floatp (nth 2 form)))))
	(setq form (list (if (eq '+ op) '1+ '1-)
			 (byte-compile-butlast form))))
       ((or (eq last -1)
	    (and (eql last -1.0)
		 (or (byte-optimize-floatp first)
		     (byte-optimize-floatp (nth 2 form)))))
	(setq form (list (if (eq '+ op) '1- '1+)
			 (byte-compile-butlast form)))))))
    form))


(defun byte-optimize-plus (form)
  (let ((orig form))
    (setq form (byte-optimize-arith-preprocess form))
    (cond
     ;; (+) -> 0
     ((null (cdr form)) (setq form 0))
     ;; (+ x) -> x, unless marker
     ((null (cddr form))
      (if (byte-optimize-nonmarker-p (cadr form))
	  (setq form (cadr form))))
     ;; otherwise
     (t
      (setq form (byte-optimize-arith-one form '+))))
    ;; binary operation
    (setq form (byte-optimize-force-binary-operation-if form 4))
    ;; for caller's sake
    (if (equal orig form)
	orig
      form)))


(defun byte-optimize-minus (form)
  (let ((orig form))
    (setq form (byte-optimize-arith-preprocess form))
    (cond
     ;; (-) --> 0
     ((null (cdr form)) (setq form 0))
     ;; (- x)
     ((null (cddr form))
      (if (numberp (nth 1 form))
	  (setq form (- (nth 1 form)))))
     ;; (- x 0)
     ((and (= 3 (length form))
	   (or (and (eq (nth 2 form) 0)
		    (byte-optimize-nonmarker-p (nth 1 form)))
	       (and (eql (nth 2 form) 0.0)
		    (byte-optimize-floatp (nth 1 form)))))
      (setq form (nth 1 form)))
     ;; otherwise
     (t
      (setq form (byte-optimize-arith-one form '-))))
    ;; binary operation
    (setq form (byte-optimize-force-binary-operation-if form 4))
    ;; for caller's sake
    (if (equal orig form)
	orig
      form)))


(defun byte-optimize-multiply (form)
  (let ((orig form))
    (setq form (byte-optimize-arith-preprocess form))
    (cond
     ;; (*)
     ((null (cdr form)) (setq form 1))
     ;; (* x)
     ((null (cddr form))
      (if (byte-optimize-nonmarker-p (cadr form))
	  (setq form (cadr form))))
     ;; (* ... 0 ...) --> (progn .. 0.0)
     ((or (memql 0.0 (cdr form))
	  (and (memq 0 (cdr form))
	       (byte-optimize-args-float-p (cdr form))))
      (setq form
	    (nconc (list 'progn)
		   (delq t (mapcar
			    (lambda (x)
				(or (byte-optimize-side-effect-free-p x)
				    x))
			      (cdr form)))
		   (list 0.0))))
     ;; (* ... 0 ...) --> (progn .. 0) or (* 0 ..)
     ((memq 0 (cdr form))
      (setq form
	    (if (byte-optimize-args-integer-or-marker-p (cdr form))
		(nconc (list 'progn)
		       (delq t (mapcar
				(lambda (x)
				  (or (byte-optimize-side-effect-free-p x)
				      x))
				(cdr form)))
		       (list 0))
	      (nconc (list (car form) 0)
		     (delq t (mapcar
			      (lambda (x) (or (integerp x) x))
			      (cdr form))))))))
    ;; binary operation
    (setq form (byte-optimize-force-binary-operation-if form 4))
    ;; for caller's sake
    (if (equal orig form)
	orig
      form)))


(defun byte-optimize-divide (form)
  (if (or (< (length form) 3) (memq 0 (cddr form)) (memql 0.0 (cddr form)))
      ;; leave runtime error
      form
    (let ((orig form))
      (setq form (byte-optimize-arith-preprocess form))
      (cond
       ;; (/ 0 ...) --> (progn .. 0.0)
       ((or (eql (nth 1 form) 0.0)
	    (and (eq (nth 1 form) 0)
		 (byte-optimize-args-float-p (cddr form))))
	(setq form (nconc (list 'progn)
			  (delq t (mapcar
				   (lambda (x)
				     (or (byte-optimize-side-effect-free-p x)
					 x))
				   (cddr form)))
			  (list 0.0))))
       ;; (/ 0 ...) --> (progn .. 0) or (/ 0 ..)
       ((eq (nth 1 form) 0)
	(setq form (if (byte-optimize-args-integer-or-marker-p (cddr form))
		       (nconc (list 'progn)
			      (delq t (mapcar
				   (lambda (x)
				     (or (byte-optimize-side-effect-free-p x)
					 x))
				   (cddr form)))
			      (list 0))
		     (nconc (list (car form) 0)
			    (delq t (mapcar
				     (lambda (x) (or (integerp x) x))
				     (cddr form)))))))
       ((let ((last (car (reverse (cddr form)))))
	  (when (numberp last)
	    (cond
	     ;; (/ x y)
	     ((= (length form) 3)
	      (cond
	       ;; (/ CONST CONST)
	       ((numberp (nth 1 form))
		(condition-case nil
		    (setq form (eval form))
		  (error)))
	       ;; (/ x 1)
	       ((or (and (eq last 1)
			 (byte-optimize-nonmarker-p (nth 1 form)))
		    (and (eql last 1.0)
			 (byte-optimize-floatp (nth 1 form))))
		(setq form (nth 1 form)))
	       ;; (/ x -1) --> (- x)
	       ((or (eq last -1)
		    (and (eql last -1.0)
			 (byte-optimize-floatp (nth 1 form))))
		(setq form (list '- (nth 1 form))))))
	     ;; (/ ... -1)
	     ((or (eq last -1)
		  (and (eql last -1.0)
		       (byte-optimize-args-float-p
			(byte-compile-butlast (cdr form)))))
	      (setq form (list '- (byte-compile-butlast form)))))))))
      ;; binary operation
      (setq form (byte-optimize-force-binary-operation-if form 4))
      ;; for caller's sake
      (if (equal orig form)
	  orig
	form))))
